"""Test external backend module with missing query class."""


GRAMMAR_PREFIX = '_Z'
""":py:class:`str`: the prefix associate to this grammar, to register this backend into the general grammar.
Required by the backend auto-loader in :py:meth:`cumin.grammar.get_registered_backends`."""
